# -*- coding: utf-8 -*-
"""
 BASEmesh: a QGIS plugin for the creation of computational meshes
 for the numerical modelling software BASEMENT using Jonathan Shewchuk's mesh
 generator 'Triangle'.

 Copyright (C) 2013 Florian Hinkelammert and Christian Volz, ETH Zurich
 hinkelammert@vaw.baug.ethz.ch / volz@vaw.baug.ethz.ch

 This QGIS Plugin is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

 For the mesh generator Triangle, the licence regulations are not obvious. It's
 author Jonathan Shevchuk has the copyright on the software and want's to be
 informed if the software is used. For more information, visit
 http://www.cs.cmu.edu/~quake/triangle.html.
"""    

import math

from mesh import MESH
import CrossSection
import meshParser1D


class CROSSSECTIONCONVERSION:
    __mesh = None
    __crossSectionList = []    
    
    # class to create BASEchain "meshes", that is:
    # one cross section is an element. containing all the cross sectional information (stored in the attributes)
    # nodes contain the elevation information
    def __init__ (self):
        self.__mesh = None
        self.__crossSectionList = []    
        pass
    
    def getMesh (self):
        return self.__mesh
    
    def setMesh (self, mesh):
        self.__mesh = mesh
        pass


    def readBMG (self,  fileID, successText):
        # import .bmg file, generated by BASEchain
        try:
            bmgFile = open(fileID+'.bmg', 'r')
            # read whole file into a list and close the file
            data = bmgFile.readlines()
            bmgFile.close()
        except:
            return ("Error: could not open the file '%s.bmg'!" % fileID)
    
        parser = meshParser1D.BASEchainParser()
        success = parser.parseBASEchain(data, successText)
        if not success:
            return "parsing failed: %s" % success
        self.__determineStatTextBASEchain(parser, successText)
        self.__crossSectionList = parser.getCrossSections()
        return self.__createMeshFromCrossSections(parser.getCrossSections())
        
    def readHECRAS (self,  fileID, successText):
        # import .bmg file, generated by BASEchain
        try:
            HECRASfile = open(fileID, 'r')
            # read whole file into a list and close the file
            data = HECRASfile.readlines()
            HECRASfile.close()
        except:
            return ("Error: could not open the file '%s'!" % fileID)
    
        parser = meshParser1D.HECRASParser()
        success = parser.parseHECRAS(data, successText)
        if not success:
            for line in successText:
                success += line
            return "parsing failed:\n%s" % success
        self.__determineStatTextHECRAS(parser, successText)
        self.__crossSectionList = parser.getCrossSections()
        return self.__createMeshFromCrossSections(parser.getCrossSections())
        
    def __createMeshFromCrossSections(self, crossSectionsList):
        nodeCounter = 1
        elementCounter = 1
        self.__mesh = MESH()
        self.__mesh.newAttribute('name')
        self.__mesh.newAttribute('distance')
        for crossSection in crossSectionsList:
            nodeIDs = []
            # create nodes
            if  crossSection.isGeoReferenced():
                for x, y, z in crossSection.getGlobalPoints():
                    self.__mesh.addNode( nodeCounter, x, y, z )
                    nodeIDs.append(nodeCounter)
                    nodeCounter = nodeCounter+1
            else:
                for x,y,z in crossSection.getPoints():
                    self.__mesh.addNode( nodeCounter, x, y, z )
                    nodeIDs.append(nodeCounter)
                    nodeCounter = nodeCounter+1
            # create element
            self.__mesh.addElement( elementCounter, nodeIDs )
            # define attributes
            self.__mesh.setAttr( elementCounter, ('name', crossSection.getName()) )
            self.__mesh.setAttr( elementCounter, ('distance', crossSection.getDistance()) )
            self.__mesh.getElement(elementCounter).__crossSection = crossSection
            elementCounter = elementCounter + 1  
        # return success
        success = self.__mesh.init()
        if success!=True:
            return success
        else:
            return True
    
    def writeBMG(self, fileName, progressBar=None):
        if len(self.__crossSectionList) <= 0:
            return "error, there are no cross sections present. "
        parser = meshParser1D.BASEchainParser()
        # write bmg-file
        fobj = open("%s.bmg" % fileName,"w")
        try:
            for cs in  self.__crossSectionList:
                parser.writeCrossSectionToBasement(fobj, cs)
        finally:
            fobj.close()
        # write additional text-file with cross-section names
        # this eases the preparation of the command-file .bmc
        fobj = open("%s_cross_section_names.txt" % fileName, "w")
        try:	
            counter = 1
            for cs in self.__crossSectionList:
                if counter > 10:
                    fobj.write("\r\n")
                    counter = 1
                fobj.write('"%s",\n' % cs.getName())                
#                fobj.write("%s " % cs.getName())
        finally:
            fobj.close()

        return True
        
    def writeHECRAS(self, fileName, progressBar=None):
        if len(self.__crossSectionList) <= 0:
            return "error, there are no cross sections present. "
        parser = meshParser1D.HECRASParser()
        fobj = open("%s.g01" % fileName,"w")
        try:
            parser.writeHECRASheader(fobj, self.__crossSectionList)
            for ii, cs in enumerate(self.__crossSectionList):
                parser.writeCrossSectionToHECRAS(fobj, ii, self.__crossSectionList)
        finally:
            fobj.close()
 
        return True
        
    def __determineStatTextBASEchain(self, parser, text):
        messages = []        
        for line in text:
            messages.append(line)
        text = []
        text.append("\n---------------------------------------------------")
        text.append("Conversion finished.")
        text.append("Messages:\n")
        for line in messages:
            text.append(line)
        text.append("\n---------------------------------------------------")
        text.append("A total number of %d sections were found." % parser.getNumberCrossSections())
        text.append( "A total number of %d cross sections were converted." % parser.getNumberCrossSections())
        text.append( "Have fun with BASEMENT!")
        text.append( "---------------------------------------------------")
        
    def __determineStatTextHECRAS(self, parser, text):
        messages = []        
        for line in text:
            messages.append(line)
        text.append("\n---------------------------------------------------")
        text.append("Conversion finished.")
        text.append("Messages:\n")
        for line in messages:
            text.append(line)
        text.append("\n---------------------------------------------------")
        text.append("A total number of %d sections were found." % parser.getNumberOriginalCrossSections())
        for ii,Type in enumerate(parser.listMustNotMissing):
            if parser.listIgnoredNr[ii] > 0:
                text.append( "Data '%s' is missin in %d cross sections. These were ignored. " % (Type, parser.listIgnoredNr[ii]))
        for ii,Type in enumerate(parser.listMustNotHave):
             if parser.listIgnoredNr2[ii] > 0:
                text.append( "Type '%s' is not supported. %d structures were ignored. " % (Type, parser.listIgnoredNr2[ii]))     
        text.append( "A total number of %d cross sections were converted." % parser.getNumberCrossSections())
        text.append( "Have fun with BASEMENT!")
        text.append( "---------------------------------------------------")
          
